import logging
import numpy as np
from scipy.interpolate import interp1d
from scipy.signal import hilbert
from .das_lib import DASLib


class PyDASLib(DASLib):
    def __init__(self) -> None:
        logging.log(logging.WARNING, "DAS is slow in Python...")

    def envelope(self, RF: np.ndarray, na: int, n_el: int, N: int) -> np.ndarray:
        return hilbert(RF).transpose(1, 0, 2)

    def delay_and_sum(self, envelope, del_Tx, del_Rx):
        na, h, w = del_Tx.shape
        n_el = del_Rx.shape[0]

        beamformed = np.zeros((na, h, w), dtype=np.complex128)
        samples_axis = np.arange(0, envelope.shape[-1])
        for ang in range(na):
            for el in range(n_el):
                data_function = interp1d(samples_axis,
                                         envelope[ang, el, :],
                                         bounds_error=False, fill_value=0)
                beamformed[ang, ...] += data_function(del_Tx[ang, ...] + del_Rx[el, ...])

        return beamformed



from abc import ABC, abstractmethod
import numpy as np


class DASLib(ABC):
    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    def envelope(self, RF: np.ndarray, na: int, n_el: int, N: int) -> np.ndarray:
        pass

    @abstractmethod
    def delay_and_sum(self):
        pass



import pkg_resources
import platform
import ctypes
import numpy as np

from .implementations import CalculateIn
from .das_lib import DASLib


class NativeDASLib(DASLib):
    def __init__(self, target) -> None:
        assert target.value[0] < 2
        self.load_lib(target)

        # Define the argument types for the C++ functions
        self.lib.envelope.argtypes = [
            np.ctypeslib.ndpointer(
                dtype=np.float64, ndim=3, flags='C_CONTIGUOUS'),
            np.ctypeslib.ndpointer(
                dtype=np.float64, ndim=3, flags='C_CONTIGUOUS'),
            np.ctypeslib.ndpointer(
                dtype=np.float64, ndim=3, flags='C_CONTIGUOUS'),
            np.ctypeslib.ndpointer(
                dtype=np.int32, ndim=1, flags='C_CONTIGUOUS'),
            ctypes.c_int,
            ctypes.c_int,
            ctypes.c_int,
            ctypes.c_int
        ]

        self.lib.delay_and_sum.argtypes = [
            np.ctypeslib.ndpointer(
                dtype=np.float64, ndim=3, flags='C_CONTIGUOUS'),
            np.ctypeslib.ndpointer(
                dtype=np.float64, ndim=3, flags='C_CONTIGUOUS'),
            np.ctypeslib.ndpointer(
                dtype=np.float64, ndim=3, flags='C_CONTIGUOUS'),
            np.ctypeslib.ndpointer(
                dtype=np.float64, ndim=3, flags='C_CONTIGUOUS'),
            np.ctypeslib.ndpointer(
                dtype=np.float64, ndim=3, flags='C_CONTIGUOUS'),
            np.ctypeslib.ndpointer(
                dtype=np.float64, ndim=3, flags='C_CONTIGUOUS'),
            ctypes.c_int,
            ctypes.c_int,
            ctypes.c_int,
            ctypes.c_int,
            ctypes.c_int
        ]

    def load_lib(self, target: CalculateIn) -> None:
        # Load the shared library
        os_name = platform.system()

        lib_path = pkg_resources.resource_filename(
            'fastDAS', 'bin') + '/libdas'

        if target == CalculateIn.CUDA:
            lib_path += '_cu'

        if os_name == 'Windows':
            lib_path += '.dll'

        elif os_name == 'Linux':
            lib_path += '.so'

        else:
            raise NotImplementedError("fast-DAS is only pre-compiled for Linux and Windows. "
                                      "If you want, try using the scripted code in target=CalculateIn.Python")

        if target == CalculateIn.CUDA:
            self.check_cuda(lib_path)

        else:
            self.lib = ctypes.CDLL(lib_path)

    def check_cuda(self, lib_path: str) -> None:
        """
        checks if the CUDA environment is set up correctly

        Parameters
        ----------
        lib_path : str
            path to the shared library
        """

        lib = ctypes.CDLL(lib_path)

        lib.cuda_valid.restype = None
        lib.cuda_valid.argtypes = [ctypes.POINTER(ctypes.c_bool)]
        valid = ctypes.c_bool()
        lib.cuda_valid(ctypes.byref(valid))

        if not valid:
            raise EnvironmentError("No NVIDIA GPU detected. Try updating CUDA toolkit.")
        else:
            self.lib = lib

    def envelope(self, RF: np.ndarray, na: int, n_el: int, N: int) -> np.ndarray:
        envelope_real = np.zeros((na, n_el, N), dtype=np.float64)
        envelope_imag = np.zeros_like(envelope_real)
        start_sample = np.arange(na)*N
        samples_per_element = na * N

        self.lib.envelope(RF,
                        envelope_real,
                        envelope_imag,
                        start_sample,
                        na,
                        n_el,
                        N,
                        samples_per_element)
        return envelope_real + 1j * envelope_imag

    def delay_and_sum(self, envelope, del_Tx, del_Rx):
        na, h, w = del_Tx.shape
        _, n_el, N = envelope.shape

        us_im_real = np.zeros((na, h, w), dtype=np.float64)
        us_im_imag = np.zeros_like(us_im_real)

        self.lib.delay_and_sum(
            us_im_real,
            us_im_imag,
            np.ascontiguousarray(envelope.real),
            np.ascontiguousarray(envelope.imag),
            del_Tx,
            del_Rx,
            na,
            n_el,
            N, h, w)
        return us_im_real + 1j * us_im_imag



import setuptools

__version__ = '1.2.0'

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="fast-das",                        # This is the name of the package
    version=__version__,                    # Release version
    author="talg",                          # Full name of the author
    description="Parallel beamforming wrapper for native code",
    long_description=long_description,      # Long description read from the the readme file
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),    # List of all python modules to be installed
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Operating System :: OS Independent",
    ],                                      # Information to filter the project on PyPi website
    python_requires='>=3.6',                # Minimum version requirement of the package
    py_modules=["fastDAS"],                 # Name of the python package
    package_dir={'fastDAS' : 'fastDAS'},        # Directory of the source code of the package
    install_requires=['numpy'],              # Install other dependencies if any
    package_data={'fastDAS': ['bin/*.dll', 'bin/*.so']},
    include_package_data=True,
    project_urls={
        'Source': 'https://github.com/talg2324/fast-das',  # Link to the source code
    },
)


import numpy as np


def pad_to_nearest_power_of_2(RF: np.ndarray, n_samples: int) -> np.ndarray:
    N = int(2**np.ceil(np.log2(n_samples)))
    if N > n_samples:
        pad_0 = (0, 0)
        pad_N = (0, N-n_samples)
        RF = np.pad(RF, (pad_0, pad_0, pad_N))
    return RF, N



import numpy as np
from .utils import pad_to_nearest_power_of_2
from .implementations import CalculateIn
from .py_lib import PyDASLib
from .native_lib import NativeDASLib


class DAS():
    def __init__(self, target: CalculateIn = CalculateIn.CPP) -> None:
        """
        delay-and-sum object
        this class is a high-level link to the underlying libraries executing the calculations

        Parameters
        ----------
        target : CalculateIn, optional
            target language implementatio, either cpp, cuda, or python,
            passed as an enumerable. by default CalculateIn.CPP
        """
        if target == CalculateIn.PYTHON:
            self.lib = PyDASLib()
        else:
            self.lib = NativeDASLib(target)

    def beamform(self, RF: np.ndarray, del_Tx: np.ndarray, del_Rx: np.ndarray) -> np.ndarray:
        """
        utility function combining the envelope and delay-and-sum steps
        see descriptions of those functions for more details
        """
        env = self.envelope(RF)
        return self.delay_and_sum(env, del_Tx, del_Rx)

    def envelope(self, RF: np.ndarray) -> np.ndarray:
        """
        calculate the signal envelope along the temporal axis using the hilbert transform

        Parameters
        ----------
        RF : np.ndarray
            raw RF data of shape [n_el, n_acq, n_samples]

        Returns
        -------
        np.ndarray
            complex hilbert transform of the input
            NOTE: the return value has shape [n_acq, n_el, N]
            this transposes the first two dimensions and pads n_samples to N (the next power of two)
        """
        n_el, na, n_samples = RF.shape
        RF, N = pad_to_nearest_power_of_2(RF, n_samples)
        return self.lib.envelope(RF.astype(np.float64), na, n_el, N)

    def delay_and_sum(self, envelope: np.ndarray, del_Tx: np.ndarray, del_Rx: np.ndarray) -> np.ndarray:
        """
        delay and sum algorithm implemented using linear interpolation

        Parameters
        ----------
        envelope : np.ndarray
            complex hilbert transform of the signal
            should have shape [n_acq, n_el, N]
        del_Tx : np.ndarray
            transmit delay matrix for an output of shape HxW
            should have shape [n_acq, H, W] and units [samples]
        del_Rx : np.ndarray
            receive delay matrix for an output of shape HxW
            should have shape [n_el, H, W] and units [samples]

        Returns
        -------
        np.ndarray
            complex beamformed ultrasound image beamformed for each transmission individually
            will have the same shape as the del_Tx input
        """
        return self.lib.delay_and_sum(envelope, del_Tx, del_Rx)
